/*
 * JScience - Java(TM) Tools and Libraries for the Advancement of Sciences.
 * Copyright (C) 2005 - JScience (http://jscience.org/)
 * All rights reserved.
 * 
 * Permission to use, copy, modify, and distribute this software is
 * freely granted, provided that this notice is preserved.
 */
package org.jscience.physics.quantities;
import org.jscience.physics.units.ConversionException;
import org.jscience.physics.units.SI;
import org.jscience.physics.units.Unit;

/**
 * This class represents an electric capacitance. The system unit for this
 * quantity is "s<sup>4</sup>·A²/(m²·kg)".  By default, instances of this class
 * are showed in Farad ("F").
 *
 * @author  <a href="mailto:jean-marie@dautelle.com">Jean-Marie Dautelle</a>
 * @version 1.0, October 24, 2004
 */
public class ElectricCapacitance extends Quantity {

    /**
     * Holds the system unit.
     */
    private final static Unit SYSTEM_UNIT = SI.FARAD.getDimension();

    /**
     * Holds the factory for this class.
     */
    private final static Factory FACTORY = new Factory(SYSTEM_UNIT) {
        protected Quantity newQuantity() {
            return new ElectricCapacitance();
        }
    }.useFor(SI.FARAD);

    /**
     * Represents an {@link ElectricCapacitance} amounting to nothing.
     */
    public final static ElectricCapacitance ZERO = (ElectricCapacitance) valueOf(
            0, SYSTEM_UNIT);

    /**
     * Default constructor (allows for derivation).
     */
    protected ElectricCapacitance() {
    }

    /**
     * Returns the {@link ElectricCapacitance} corresponding to the specified
     * quantity.
     *
     * @param  quantity a quantity compatible with {@link ElectricCapacitance}.
     * @return the specified quantity or a new {@link ElectricCapacitance} instance.
     * @throws ConversionException if the current model does not allow the
     *         specified quantity to be converted to {@link ElectricCapacitance}.
     */
    public static ElectricCapacitance electricCapacitanceOf(Quantity quantity) {
        return (ElectricCapacitance) FACTORY.quantity(quantity);
    }

    /**
     * Shows {@link ElectricCapacitance} instances in the specified unit.
     *
     * @param  unit the output unit for {@link ElectricCapacitance} instances.
     * @see    Quantity#getOutputUnit
     */
    public static void showAs(Unit unit) {
        FACTORY.showInstancesAs(unit);
    }

    private static final long serialVersionUID = 2225191356834086422L;
}