/*
 * JScience - Java(TM) Tools and Libraries for the Advancement of Sciences.
 * Copyright (C) 2005 - JScience (http://jscience.org/)
 * All rights reserved.
 * 
 * Permission to use, copy, modify, and distribute this software is
 * freely granted, provided that this notice is preserved.
 */
package org.jscience.physics.quantities;
import org.jscience.physics.units.ConversionException;
import org.jscience.physics.units.SI;
import org.jscience.physics.units.Unit;

/**
 * This class represents the number of times a specified phenomenon occurs
 * within a specified interval. The system unit for this quantity is "1/s".
 * By default, instances of this quantity are showed in Hertz ("Hz").
 *
 *
 * @author  <a href="mailto:jean-marie@dautelle.com">Jean-Marie Dautelle</a>
 * @version 1.0, October 24, 2004
 */
public class Frequency extends Quantity {

    /**
     * Holds the system unit.
     */
    private final static Unit SYSTEM_UNIT = SI.HERTZ.getDimension();

    /**
     * Holds the factory for this class.
     */
    private final static Factory FACTORY = new Factory(SYSTEM_UNIT) {
        protected Quantity newQuantity() {
             return new Frequency();
        }
    }.useFor(SI.HERTZ);

    /**
     * Represents a {@link Frequency} amounting to nothing.
     */
    public final static Frequency ZERO = (Frequency) valueOf(0, SYSTEM_UNIT);

    /**
     * Default constructor (allows for derivation).
     */
    protected Frequency() {}

    /**
     * Returns the {@link Frequency} corresponding to the specified quantity.
     *
     * @param  quantity a quantity compatible with {@link Frequency}.
     * @return the specified quantity or a new {@link Frequency} instance.
     * @throws ConversionException if the current model does not allow the
     *         specified quantity to be converted to {@link Frequency}.
     */
    public static Frequency frequencyOf(Quantity quantity) {
        return (Frequency) FACTORY.quantity(quantity);
    }

    /**
     * Shows {@link Frequency} instances in the specified unit.
     *
     * @param  unit the output unit for {@link Frequency} instances.
     * @see    Quantity#getOutputUnit
     */
    public static void showAs(Unit unit) {
        FACTORY.showInstancesAs(unit);
    }

    private static final long serialVersionUID = -3137044723817499548L;
}
